/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id : $ 
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.engine.eip;

import java.io.IOException;

import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.TransformerException;

import org.objectweb.petals.component.common.PEtALSComponentSDKException;
import org.objectweb.petals.component.common.util.SourceHelper;
import org.w3c.dom.Node;

/**
 * Object to store informations during the execution of the calls.
 * Contains the in message, the out message and the aggregate result.
 * @version $Rev: $ $Date: $
 * @author Guillaume Decarnin - Open Wide
 */
public class EipContext {
    
    public static enum ExecutionContext { CHAIN, DISPATCH, AGGREGATE };
    
    private static final String AGGREGATE_RESULT_TAG = "aggregate-result";
    
    private NormalizedMessage in = null;
    private NormalizedMessage out = null;
    private Node inNode = null;
    private Node outNode = null;
    private StringBuilder aggregateResult = new StringBuilder();
    private int nbOfOpenAggregate = 0;
    
    public NormalizedMessage getIn() {
        return in;
    }
    
    public Node getInNode() throws TransformerException, PEtALSComponentSDKException, IOException {
        if (inNode == null && in != null) {
            inNode = EipEngine.createDOMNodeFromSource(in.getContent());
        }
        return inNode;
    }
    
    public Node getOutNode() throws TransformerException, PEtALSComponentSDKException, IOException {
        if (outNode == null && out != null) {
            outNode = EipEngine.createDOMNodeFromSource(out.getContent());
        }
        return outNode;
    }

    public void setIn(NormalizedMessage in) {
        this.in = in;
        this.inNode = null;
    }

    public NormalizedMessage getOut() {
        return out;
    }

    public void setOut(NormalizedMessage out) {
        this.out = out;
    }
    
    /**
     * Find the execution context of a node: chain, dispatch, aggregate or null
     * @param node The node
     * @return The execution context
     */
    public ExecutionContext getExecutionContext(Node node) {
        Node parent = node.getParentNode();
        
        while (parent != null) {
            String name = parent.getNodeName();
            
            if (name == null) {
                return null;
            }
            
            if (name.equals("chain")) {
                return ExecutionContext.CHAIN;
            }
            else if (name.equals("dispatch")) {
                return ExecutionContext.DISPATCH;
            }
            else if (name.equals("aggregate")) {
                return ExecutionContext.AGGREGATE;
            }
            else if (name.equals("extension")) {
                return null;
            }
            
            parent = parent.getParentNode();
        }
        
        return null;
    }
    
    /**
     * Add the content of the current out to the aggregate result.
     * @throws PEtALSComponentSDKException
     */
    public void addOutContentToAggregateResult() throws PEtALSComponentSDKException {
        String result = SourceHelper.createString(out.getContent());
        aggregateResult.append(result);
        aggregateResult.append("\n");
    }
    
    /**
     * Notify an open aggregate tag.
     */
    public void enterAggregate() {
        nbOfOpenAggregate++;
        
        aggregateResult.append("<");
        aggregateResult.append(AGGREGATE_RESULT_TAG);
        aggregateResult.append(">\n");
    }

    /**
     * Notify a closed aggregate tag.
     * @param msgExchange The MessageExchange to create the out message.
     * @throws MessagingException
     * @throws PEtALSComponentSDKException
     */
    public void exitAggregate(MessageExchange msgExchange) throws MessagingException, PEtALSComponentSDKException {
        nbOfOpenAggregate--;
        
        aggregateResult.append("</");
        aggregateResult.append(AGGREGATE_RESULT_TAG);
        aggregateResult.append(">\n");
        
        if (nbOfOpenAggregate == 0) {
            // the last aggragate tag is closed
            // write aggregate result and clear string
            String result = aggregateResult.toString();
            
            out = msgExchange.createMessage();
            out.setContent(SourceHelper.createSource(result));
            
            aggregateResult.delete(0, aggregateResult.length());
        }
    }
}
