/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id : $ 
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.engine.eip;

import java.io.IOException;
import java.io.StringWriter;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;

import javax.activation.DataHandler;
import javax.jbi.JBIException;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.InOnly;
import javax.jbi.messaging.InOptionalOut;
import javax.jbi.messaging.InOut;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.messaging.RobustInOnly;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.objectweb.petals.common.descriptor.Extensions;
import org.objectweb.petals.common.descriptor.UnknownExtension;
import org.objectweb.petals.common.util.XMLUtil;
import org.objectweb.petals.component.common.HandlingException;
import org.objectweb.petals.component.common.PEtALSComponentSDKException;
import org.objectweb.petals.component.common.se.AbstractServiceEngine;
import org.objectweb.petals.component.common.util.MessageExchangeWrapper;
import org.objectweb.petals.component.common.util.SourceHelper;
import org.objectweb.petals.engine.eip.EipContext.ExecutionContext;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * <h1>Enterprise Integration Patterns Engine</h1>
 * 
 * <h1>1. Purpose</h1>
 * 
 * <p>This component allows to implement some Enterprise Integration Patterns.
 * The fonctionnalities are:</p>
 * <ul>
 * <li>chain services calls</li>
 * <li>dispatch messages</li>
 * <li>aggregate the results of services calls</li>
 * <li>dynamic orchestration with conditions, recipients list in message and integration of routing services</li>
 * </ul>
 * 
 * 
 * <h1>2. Operations</h1>
 * 
 * <h2>process</h2>
 * <p>Process the orchestation using the current Service Unit configuration.</p>
 * 
 * <h2>getEipExtension</h2>
 * <p>Return the XML representation of the orchestration.</p>
 * 
 * <h2>setEipExtension</h2>
 * <p>Set a new Service Unit configuration. The content of the message is the new "extension" node
 * to use (same as in jbi.xml).</p> 
 * 
 * 
 * <h1>3. Orchestation format</h1>
 * 
 * <p>The orchestration process is writen in the Service Unit configuration file "jbi.xml".</p>
 * 
 * <p>The XML format is described in the file "src/main/jbi/META-INF/eip-extension.xsd".
 * This schema is used to validate the configuration on Service Unit deployement and during the setEipExtension operation.</p>
 * 
 * 
 * <h2>The "extension" element</h2>
 * <p>The root element "extension" uses the namespace "http://petals.objectweb.org/extensions/eip/".</p>
 * <p>The root must contain one child : "chain", "dispatch" or "aggregate".</p>
 * 
 * 
 * <h2>The "call" element</h2>
 * <p>The "call" element invokes a service. The attribute "method" is optional.
 * Its default value is "InOut".</p>
 * 
 * <pre>&lt;call service="{http://petals.objectweb.org/}SampleTestService"
 *      operation="myOperation"
 *      method="InOnly|InOut|InOptionalOut|RobustInOnly"/&gt;</pre>
 * 
 * <h2>The execution context</h2>
 * <p>There are 3 execution contexts : "chain", "dispatch" and "aggregate".
 * These elements define the transitions between the calls.</p>
 * 
 * <h3>The "chain" element</h3>
 * <p>In a "chain" element, the out message of a call is the in message of the next call.
 * After the "chain" bloc, the current message is the result of the last call.</p>
 * 
 * <pre>&lt;chain&gt;
 *    &lt;call service="service1" operation="operation1"/&gt;
 *    &lt;call service="service2" operation="operation2"/&gt;
 *    ...
 *&lt;/chain&gt;</pre>
 * 
 * <h3>The "dispatch" element</h3>
 * <p>In a "dispatch" element, the same message is send to all the services. The order is not important.
 * After a dispatch, the current message is the same as the initial message.</p>
 * 
 * <pre>&lt;dispatch&gt;
 *    &lt;call service="service1" operation="operation1"/&gt;
 *    &lt;call service="service2" operation="operation2"/&gt;
 *    ...
 *&lt;/dispatch&gt;</pre>
 * 
 * <h3>The "aggregate" element</h3>
 * <p>In an "aggregate" element, the same message is send to all the services.
 * After an aggregate, the current message is an aggregation of all the results of the inside calls,
 * surrounded with a "aggregate-result" tag.</p>
 * 
 * <pre>&lt;aggregate&gt;
 *    &lt;call service="service1" operation="operation1"/&gt;
 *    &lt;call service="service2" operation="operation2"/&gt;
 *    ...
 *&lt;/aggregate&gt;</pre>
 *
 * <p>Result of an aggregate with a service who returns a number:</p>
 * 
 * <pre>&lt;aggregate-result&gt;
 *    &lt;number&gt;3&lt;/number&gt;
 *    &lt;number&gt;4&lt;/number&gt;
 *    &lt;number&gt;6&lt;/number&gt;
 *&lt;/aggregate-result&gt;</pre>
 * 
 * 
 * <h2>Dynamic calls</h2>
 * 
 * <p>The EIP component can generate a list of calls and execute them.</p>
 * 
 * <h3>The "get-calls-by-service" element</h3>
 * <p>This element calls a service who returns a list of calls.</p>
 * 
 * <pre>&lt;get-calls-by-service 
 *        service="{http://petals.objectweb.org/}SampleTestService"
 *        operation="getCalls"/&gt;</pre>
 * 
 * <p>The operation "getCalls" have to return a message like this:</p>
 * 
 * <pre>&lt;calls&gt;
 *    &lt;call service="myService" operation="myOperation"/&gt;
 *    &lt;call service="myService2" operation="myOperation2"/&gt;
 *    ...
 *&lt;/calls&gt;</pre>
 * 
 * <h3>The "get-calls-by-xpath" element</h3>
 * <p>This element allows to read the routing path included in the message.</p>
 * 
 * <pre>&lt;get-calls-by-xpath
 *        base="/message/recipients/recipient"
 *        service="@service" operation="@operation"/&gt;</pre>
 * 
 * <p>The concatenation of the attributes "base" and "service" is a XPath, used to extract a
 * list of services from the message. "base" and "operation" extract the operations.</p>
 * 
 * <p>Sample of message who respects the previous XPath:</p>
 * 
 * <pre>&lt;message&gt;
 *    &lt;recipients&gt;
 *        &lt;recipient service="service1" operation="operation1"/&gt;
 *        &lt;recipient service="service2" operation="operation2"/&gt;
 *        ...
 *    &lt;/recipients&gt;
 *    &lt;content&gt;
 *        ...
 *    &lt;/content&gt;
 *&lt;/message&gt;</pre>
 * 
 * 
 * <h2>Conditions</h2>
 * 
 * <h3>The "if" element</h3>
 * <p>The "if" node has a "test" attibute who contains a XPath boolean expression. The condition is
 * evaluated using the message content. If the condition returns "true", then the bloc inside the "if" is executed.</p>
 * 
 * <pre>&lt;if test="xpathBooleanExpression"&gt;
 *    ...
 *&lt;/if&gt;</pre>
 * 
 * <h3>The "choose" element</h3>
 * <p>The "choose" element has two children: "when" and "otherwise".
 * "when" has a "test" attribute, like the "if" element. If the condition is true,
 * the "when" bloc is executed, else the "otherwise" bloc is executed.</p>
 * 
 * <pre>&lt;choose&gt;
 *    &lt;when test="xpathBooleanExpression"&gt;
 *        ...
 *    &lt;/when&gt;
 *    &lt;otherwise&gt;
 *        ...
 *    &lt;/otherwise&gt;
 *&lt;/choose&gt;</pre>
 * 
 * 
 * <h2>Imbrications</h2>
 * 
 * <p>It is possible to imbricate the elements. Any bloc can contain these elements:</p>
 * <ul>
 * <li>call</li>
 * <li>get-calls-by-service</li>
 * <li>get-calls-by-xpath</li>
 * <li>if</li>
 * <li>choose</li>
 * <li>chain</li>
 * <li>dispatch</li>
 * <li>aggregate</li>
 * </ul>
 * 
 * @version $Rev: $ $Date: $
 * @author Marc Dutoo - Open Wide
 * @author Guillaume Decarnin - Open Wide
 */
public class EipEngine extends AbstractServiceEngine {

    private static final String PROCESS_OPERATION = "process";
    private static final String GET_EIP_EXTENSION_OPERATION = "getEipExtension";
    private static final String SET_EIP_EXTENSION_OPERATION = "setEipExtension";
    
    // TODO use org.objectweb.petals.component.common.sender.ROBUST_IN_ONLY
    private static final String INONLY_METHOD = "InOnly";
    private static final String INOUT_METHOD = "InOut";
    private static final String INOPTIONALOUT_METHOD = "InOptionalOut";
    private static final String ROBUSTINONLY_METHOD = "RobustInOnly";
    
    private static final String EIP_NAMESPACE_STRING = "http://petals.objectweb.org/extensions/eip/";
    public  static URI EIP_NAMESPACE_URI;
    public  static final XPath xpath;
    
    private MessageExchangeFactory mef = null;
       
    static {
        // init static fields
        xpath = XPathFactory.newInstance().newXPath();
        xpath.setNamespaceContext(new CustomNamespaceContext("eip", EIP_NAMESPACE_STRING));
        
        try {
            EIP_NAMESPACE_URI = new URI(EIP_NAMESPACE_STRING);
        } catch (URISyntaxException e) {
            e.printStackTrace();
        }
    }
    
    @Override
    protected void init() throws JBIException {
        setServiceUnitExtensionsValidator(new EipServiceUnitExtensionsValidator(getContext().getInstallRoot()));
    }
    
    @Override
    public void start() throws JBIException {
        mef = getChannel().createExchangeFactory();
        super.start();
    }
    
    @Override
    public void stop() throws JBIException {
        mef = null;
        super.stop();
    }
    
    /**
     * Extract the EIP extension node from unknownsExtensions map.
     * @param exchange
     * @param extensions
     * @return The EIP extension node
     * @throws HandlingException
     */
    protected Node getExtensionNode(MessageExchangeWrapper exchange, Extensions extensions) throws HandlingException {
        UnknownExtension eipExtension = extensions.getUnknownsExtensions().get(EIP_NAMESPACE_URI);
        if (eipExtension == null) {
            throw new HandlingException("The EIP extension node is null.");
        }
        return eipExtension.getNode();
    }
    
    @Override
    protected boolean onExchange(MessageExchangeWrapper exchange, Extensions extensions) throws HandlingException {
        String operationName = exchange.getOperationName();
        
        try {
            if (PROCESS_OPERATION.equals(operationName)) {
                return processOperation(exchange, extensions);
            }
            
            if (GET_EIP_EXTENSION_OPERATION.equals(operationName)) {
                return getEipExtensionOperation(exchange, extensions);
            }
            
            if (SET_EIP_EXTENSION_OPERATION.equals(operationName)) {
                return setEipExtensionOperation(exchange, extensions);
            }
        } catch (Exception e) {
            throw new HandlingException(e);
        }
        
        throw new HandlingException("Unknown operation: \"" + operationName + "\".");
    }
    
    /**
     * Write the EIP extension node in the out message.
     * @param exchange
     * @param extensions
     * @return
     * @throws Exception
     */
    protected boolean getEipExtensionOperation(MessageExchangeWrapper exchange, Extensions extensions) throws Exception {
        // find extension node
        Node extensionNode = getExtensionNode(exchange, extensions);
        
        // write extension in out message 
        exchange.setOutMessageContent(XMLUtil.createStringFromDOMNode(extensionNode, true));
        
        return true;
    }
    
    /**
     * Extract the EIP extension node from in message and update the SU configuration.
     * @param exchange
     * @param extensions
     * @return
     * @throws Exception
     */
    protected boolean setEipExtensionOperation(MessageExchangeWrapper exchange, Extensions extensions) throws Exception {
        // extract extension node from message
        Node extensionNode = createDOMNodeFromSource(exchange.getInMessage().getContent()).getFirstChild();
        
        // validate extension node
        ((EipServiceUnitExtensionsValidator) getServiceUnitExtensionsValidator()).validateExtensionNode(extensionNode);
        
        // update extension map
        UnknownExtension unknownExtension = new UnknownExtension();
        unknownExtension.setNode(extensionNode);
        extensions.getUnknownsExtensions().put(EIP_NAMESPACE_URI, unknownExtension);
        
        // return ok
        exchange.setOutMessageContent("<success/>");
        
        return true;
    }
    
    /**
     * Process orchestration defined in SU configuration.
     * @param exchange
     * @param extensions
     * @return
     * @throws Exception
     */
    protected boolean processOperation(MessageExchangeWrapper exchange, Extensions extensions) throws Exception {
        // find extension node
        Node extensionNode = getExtensionNode(exchange, extensions);
        
        // init eipContext
        EipContext eipContext = new EipContext();
        eipContext.setIn(exchange.getInMessage());
        
        // start orchestration
        evaluateChildren(extensionNode, eipContext);
        
        // after orchestration
        if (eipContext.getOut() == null) {
            // last forward was InOnly, so let's return "<success/>"
            exchange.setOutMessageContent("<success/>");
        } else {
            // let's return the result of the last forward
            copy(eipContext.getOut(), exchange);
        }
        
        return true;
    }
    
    /**
     * Evaluate child nodes of the node parameter.
     * @param node
     * @param eipContext
     * @throws Exception
     * @throws EipConfigurationException
     * @throws XPathExpressionException
     */
    protected void evaluateChildren(Node node, EipContext eipContext) throws Exception, EipConfigurationException, XPathExpressionException {
        NodeList children = node.getChildNodes();
        
        for (int i = 0; i < children.getLength(); i++) {
            Node item = children.item(i);
            if (item.getNodeType() != Node.ELEMENT_NODE) {
                continue;
            }
            
            String name = item.getNodeName();
            if (name.equals("call")) {
                evaluateCallNode(item, eipContext);
                
                // preparing next call
                ExecutionContext executionContext = eipContext.getExecutionContext(item);
                if (executionContext == ExecutionContext.CHAIN) {
                    // the next in message is the current out message
                    eipContext.setIn(eipContext.getOut());
                }
                else if (executionContext == ExecutionContext.AGGREGATE) {
                    // add result to the aggregate String
                    eipContext.addOutContentToAggregateResult();
                }
            }
            
            else if (name.equals("get-calls-by-service")) {
                evaluateGetCallsByServiceNode(item, eipContext);
                
                // update children after the insertion of call nodes
                children = node.getChildNodes();
            }
            
            else if (name.equals("get-calls-by-xpath")) {
                evaluateGetCallsByXPathNode(item, eipContext);
                
                // update children after the insertion of call nodes
                children = node.getChildNodes();
            }
            
            else if (name.equals("if")) {
                evaluateIfNode(item, eipContext);
            }
            
            else if (name.equals("choose")) {
                evaluateChooseNode(item, eipContext);
            }
            
            else if (name.equals("chain")) {
                NormalizedMessage in = eipContext.getIn();
                
                evaluateChildren(item, eipContext);
                
                ExecutionContext executionContext = eipContext.getExecutionContext(item);
                if (executionContext == ExecutionContext.AGGREGATE) {
                    // we are in an aggregate after this chain
                    
                    // add result to the aggregate String
                    eipContext.addOutContentToAggregateResult();
                }
                if (executionContext == ExecutionContext.AGGREGATE || executionContext == ExecutionContext.DISPATCH) {
                    // we are in a parallel context after this chain
                    
                    // set the right in message
                    eipContext.setIn(in);
                }
            }
            
            else if (name.equals("dispatch")) {
                evaluateChildren(item, eipContext);
            }
            
            else if (name.equals("aggregate")) {
                eipContext.enterAggregate();
                
                evaluateChildren(item, eipContext);
                
                eipContext.exitAggregate(createMessageExchange(INOUT_METHOD));
                
                if (eipContext.getExecutionContext(item) == ExecutionContext.CHAIN) {
                    // we are in a chain after this aggregate
                    
                    // set the right in message
                    eipContext.setIn(eipContext.getOut());
                }
            }
        }
    }
    
    /**
     * Create a MessageExchange with the MessageExchangeFactory.
     * @param method
     * @return
     * @throws Exception
     */
    private MessageExchange createMessageExchange(String method) throws Exception {
        MessageExchange msg;
        if (INONLY_METHOD.equals(method)) {
            msg = mef.createInOnlyExchange();
        } else if (INOUT_METHOD.equals(method)) {
            msg = mef.createInOutExchange();
        } else if (INOPTIONALOUT_METHOD.equals(method)) {
            msg = mef.createInOptionalOutExchange();
        } else if (ROBUSTINONLY_METHOD.equals(method)) {
            msg = mef.createRobustInOnlyExchange();
        } else {
            throw new Exception("Unknown method " + method);
        }
        return msg;
    }
    
    /**
     * Evaluate the XPath condition of a node (attribute "test" in a "if" node or a "when" node).
     * @param node The "if" node or the "when" node to evaluate.
     * @param message The condition is tested with the message content.
     * @return The result of condition evaluation.
     * @throws XPathExpressionException
     */
    private boolean evaluateCondition(Node node, Node message) throws XPathExpressionException {
        String condition = null;
        String debugPath = getDebugPath(node);
        
        try {
            condition = node.getAttributes().getNamedItem("test").getTextContent();
            XPathExpression expression = xpath.compile(condition);
            String conditionResult = expression.evaluate(message);
            
            getLogger().log(Level.FINE, "Evaluate condition: \"" + condition + "\", result: \"" + conditionResult + "\", node: \"" + debugPath + "\".");
            
            return conditionResult.equals("true");
            
        } catch (XPathExpressionException e) {
            String err = "Error in XPathExpression \"" + condition + "\", cause: \"" + e.getCause().getMessage() + "\", node: \"" + debugPath + "\".";
            getLogger().log(Level.SEVERE, err);
            throw e;
        }
    }
    
    /**
     * Evaluate children of a "if" node if the condition is true.
     * @param ifNode
     * @param eipContext
     * @throws Exception
     * @throws EipConfigurationException
     * @throws XPathExpressionException
     */
    private void evaluateIfNode(Node ifNode, EipContext eipContext) throws Exception, EipConfigurationException, XPathExpressionException {
        if (evaluateCondition(ifNode, eipContext.getInNode())) {
            evaluateChildren(ifNode, eipContext);
        }
    }
    
    /**
     * Test the condition and evaluate the "when" bloc or the "otherwise" bloc.
     * @param chooseNode
     * @param eipContext
     * @throws Exception
     * @throws EipConfigurationException
     * @throws XPathExpressionException
     */
    private void evaluateChooseNode(Node chooseNode, EipContext eipContext) throws Exception, EipConfigurationException, XPathExpressionException {
        // evaluate condition
        Node when = XMLUtil.findChild(chooseNode, "when", false);
        if (when == null) {
            String err = "The content of element 'choose' is not complete. 'when' is expected.";
            getLogger().log(Level.SEVERE, err); 
            throw new EipConfigurationException(err);
        }
        boolean test = evaluateCondition(when, eipContext.getInNode());
        
        // choose the bloc to execute
        if (test) {
            // when
            evaluateChildren(when, eipContext);
        } else {
            // otherwise
            Node otherwise = XMLUtil.findChild(chooseNode, "otherwise", false);
            if (otherwise == null) {
                String err = "The content of element 'choose' is not complete. 'otherwise' is expected.";
                getLogger().log(Level.SEVERE, err); 
                throw new EipConfigurationException(err);
            }
            evaluateChildren(otherwise, eipContext);
        }
    }
    
    /**
     * Call the service specified in the "call" node.
     * @param callNode
     * @param eipContext
     * @throws Exception
     */
    private void evaluateCallNode(Node callNode, EipContext eipContext) throws Exception {
        try {
            String targetServiceName = callNode.getAttributes().getNamedItem("service").getTextContent();
            String targetServiceOperation = callNode.getAttributes().getNamedItem("operation").getTextContent();
            
            String targetServiceMethod = INOUT_METHOD;
            Node targetServiceMethodItem = callNode.getAttributes().getNamedItem("method");
            if (targetServiceMethodItem != null) {
                targetServiceMethod = targetServiceMethodItem.getTextContent();
            }
            
            getLogger().log(Level.FINE, "Call service: \"" + targetServiceName + "\", operation: \"" + targetServiceOperation + "\", method: \"" + targetServiceMethod + "\", node: \"" + getDebugPath(callNode) + "\".");
            
            // getting available endpoints
            QName targetService = QName.valueOf(targetServiceName);
            ServiceEndpoint[] targetEndpoints = getContext().getEndpointsForService(targetService);
            if (targetEndpoints == null || targetEndpoints.length == 0) {
                throw new Exception("No endpoint found for service " + targetServiceName);
            }

            if (eipContext.getIn() == null) {
                throw new Exception("Previous forward had no OUT message "
                    + "so there can't be another forward afterwards to "
                    + targetServiceName + "." + targetServiceOperation);
            }
            
            try {
                NormalizedMessage currentOut = forwardCopy(eipContext.getIn(), targetService, targetServiceOperation, targetServiceMethod);
                eipContext.setOut(currentOut);
            } catch (Exception e) {
                throw new Exception("Error when forwarding to service and operation "
                    + targetServiceName + "." + targetServiceOperation, e);
            }
            if (eipContext.getOut() == null) {
                // the out message was already treated
                // NB. we'll check that there will be no more forward
                // and return "<success/>"
                return;

            } else if (eipContext.getOut() instanceof Fault) {
                String faultString = SourceHelper.createString(eipContext.getOut().getContent());
                throw new Exception("Service fault received from call "
                    + "to service and operation " + targetServiceName
                    + "." + targetServiceOperation + " : " + faultString);
            }
            
        } catch (XPathExpressionException e) {
            e.printStackTrace();
        }
    }
    
    /**
     * Get the list of "call" nodes and insert each service returned in the orchestration tree. 
     * Call the service specified in the "call" node.
     * @param getCallsByServiceNode
     * @param eipContext
     * @throws Exception
     */
    private void evaluateGetCallsByServiceNode(Node getCallsByServiceNode, EipContext eipContext) throws Exception {
        // call service
        evaluateCallNode(getCallsByServiceNode, eipContext);
        
        // the result is a list of callNode
        Node calls = eipContext.getOutNode().getFirstChild();

        if (calls == null || calls.getNodeType() != Node.ELEMENT_NODE) {
            throw new Exception("Expected <calls> encapsulated content but found a empty document.");
        }
        if (!calls.getNodeName().equals("calls")) {
            throw new Exception("Expected <calls> encapsulated content but found <" + calls.getNodeName() + ">.");
        }
        
        // find the insertion point
        Node nextSibling = getNextSibling(getCallsByServiceNode);
        Node parent = getCallsByServiceNode.getParentNode();
        Document doc = getCallsByServiceNode.getOwnerDocument();
        
        // insert calls in the main document
        int nbInserts = 0;
        NodeList children = calls.getChildNodes();
        for (int i = 0; i < children.getLength(); i++) {
            Node child = children.item(i);
            if (child.getNodeType() != Node.ELEMENT_NODE) {
                continue;
            }

            // validation
            if (!child.getNodeName().equals("call")) {
                throw new Exception("Invalid content was found : <" + child.getNodeName() + ">. Only <call> is allowed.");
            }
            
            Node serviceNode = child.getAttributes().getNamedItem("service");
            Node operationNode = child.getAttributes().getNamedItem("operation");
            Node methodNode = child.getAttributes().getNamedItem("method");
            
            if (serviceNode == null) {
                throw new Exception("Attribute 'service' must appear on element 'call'.");
            }
            if (operationNode == null) {
                throw new Exception("Attribute 'operation' must appear on element 'call'.");
            }
            
            // create a copy node
            Node call = doc.createElement("call");

            // copy node values
            Attr service = doc.createAttribute("service");
            service.setTextContent(serviceNode.getTextContent());
            call.getAttributes().setNamedItem(service);
            
            Attr operation = doc.createAttribute("operation");
            operation.setTextContent(operationNode.getTextContent());
            call.getAttributes().setNamedItem(operation);
            
            if (methodNode != null) {
                Attr method = doc.createAttribute("method");
                method.setTextContent(methodNode.getTextContent());
                call.getAttributes().setNamedItem(method);
            }
            
            // insert node
            parent.insertBefore(call, nextSibling);
            nbInserts++;
        }
        
        getLogger().log(Level.FINE, "get-calls-by-service: insert " + nbInserts + " call(s).");
    }
    
    /**
     * Get the list of "call" nodes and insert each service returned in the orchestration tree.
     * @param getCallsByXPathNode
     * @param eipContext
     * @throws Exception
     */
    private void evaluateGetCallsByXPathNode(Node getCallsByXPathNode, EipContext eipContext) throws Exception {
        String base = getCallsByXPathNode.getAttributes().getNamedItem("base").getTextContent();
        String service = getCallsByXPathNode.getAttributes().getNamedItem("service").getTextContent();
        String operation = getCallsByXPathNode.getAttributes().getNamedItem("operation").getTextContent();
        String method = null;
        if (getCallsByXPathNode.getAttributes().getNamedItem("method") != null) {
            method = getCallsByXPathNode.getAttributes().getNamedItem("method").getTextContent();
        }
        
        // add a slash between part 1 and part 2 of the XPath expressions
        if (!base.endsWith("/")) {
            if (!service.startsWith("/")) {
                service = "/" + service;
            }
            if (!operation.startsWith("/")) {
                operation = "/" + operation;
            }
            if (method != null && !method.startsWith("/")) {
                method = "/" + method;
            }
        }
        String xpathServices = base + service;
        String xpathOperations = base + operation;

        XPathExpression expressionServices = xpath.compile(xpathServices);
        XPathExpression expressionOperations = xpath.compile(xpathOperations);
        
        // extract data
        Node message = eipContext.getInNode();
        NodeList serviceNodes = (NodeList) expressionServices.evaluate(message, XPathConstants.NODESET);
        NodeList operationNodes = (NodeList) expressionOperations.evaluate(message, XPathConstants.NODESET);
        
        NodeList methodNodes = null;
        if (method != null) {
            String xpathMethods = base + method;
            XPathExpression expressionMethods = xpath.compile(xpathMethods);
            methodNodes = (NodeList) expressionMethods.evaluate(message, XPathConstants.NODESET);
        }
        
        // validation
        if (serviceNodes == null) {
            throw new Exception("The list of service nodes is null.");
        }
        if (operationNodes == null) {
            throw new Exception("The list of operation nodes is null.");
        }
        
        if (serviceNodes.getLength() != operationNodes.getLength()) {
            String err = "Found " +
                    serviceNodes.getLength() + " services node(s) and " + 
                    operationNodes.getLength() + " operation node(s) : numbers should be equals.";
            getLogger().log(Level.SEVERE, err);
            throw new Exception(err);
        }
        
        if (methodNodes != null && serviceNodes.getLength() != methodNodes.getLength()) {
            String err = "Found " +
                    serviceNodes.getLength() + " services node(s) and " + 
                    methodNodes.getLength() + " method node(s) : numbers should be equals.";
            getLogger().log(Level.SEVERE, err);
            throw new Exception(err);
        }
        
        // find the insertion point
        Node nextSibling = getNextSibling(getCallsByXPathNode);
        Node parent = getCallsByXPathNode.getParentNode();
        Document doc = getCallsByXPathNode.getOwnerDocument();
        
        // create and insert call nodes
        int nbInserts = 0;
        for (int i = 0; i < serviceNodes.getLength(); i++) {
            Node serviceNode = serviceNodes.item(i);
            Node operationNode = operationNodes.item(i);
            
            Node methodNode = null;
            if (methodNodes != null) {
                methodNode = methodNodes.item(i);
            }
            
            // create a copy node
            Node call = doc.createElement("call");
            
            // copy node values
            Attr serviceAttr = doc.createAttribute("service");
            serviceAttr.setTextContent(serviceNode.getTextContent());
            call.getAttributes().setNamedItem(serviceAttr);
            
            Attr operationAttr = doc.createAttribute("operation");
            operationAttr.setTextContent(operationNode.getTextContent());
            call.getAttributes().setNamedItem(operationAttr);
            
            if (methodNode != null) {
                Attr methodAttr = doc.createAttribute("method");
                methodAttr.setTextContent(methodNode.getTextContent());
                call.getAttributes().setNamedItem(methodAttr);
            }
            
            // insert node
            parent.insertBefore(call, nextSibling);
            nbInserts++;
        }
        
        getLogger().log(Level.FINE, "get-calls-by-xpath: insert " + nbInserts + " call(s).");
    }
    
    /**
     * Forwards a copy of the given message to the given service
     * using the given operation and method (i.e. InOut etc.)
     * @param originalIn
     * @param service
     * @param operation
     * @param method
     * @return
     * @throws Exception
     */
    @SuppressWarnings("unchecked")
    protected NormalizedMessage forwardCopy(NormalizedMessage originalIn,
                QName service, String operation, String method) throws Exception {
        // creating message exchange
        MessageExchange msg = createMessageExchange(method);
        
        // setting target
        msg.setService(service);
        msg.setOperation(QName.valueOf(operation));
        
        // copying IN message
        NormalizedMessage copyIn = msg.createMessage();
        copy(originalIn, copyIn);
        if (INONLY_METHOD.equals(method)) {
            ((InOnly) msg).setInMessage(copyIn);
        } else if (INOUT_METHOD.equals(method)) {
            ((InOut) msg).setInMessage(copyIn);
        } else if (INOPTIONALOUT_METHOD.equals(method)) {
            ((InOptionalOut) msg).setInMessage(copyIn);
        } else if (ROBUSTINONLY_METHOD.equals(method)) {
            ((RobustInOnly) msg).setInMessage(copyIn);
        } else {
            throw new Exception("Unknown method " + method);
        }
        
        // setting properties
        Set<String> propertyNames = originalIn.getPropertyNames();
        for (String propertyName : propertyNames) {
            msg.setProperty(propertyName, originalIn.getProperty(propertyName));
        }
        
        // doing the actual message send
        boolean ok = getChannel().sendSync(msg, 0);
        if(ok) {
            if (ExchangeStatus.DONE.equals(msg.getStatus())) {
                return null; // means DONE status received
                
            } else if (ExchangeStatus.ERROR.equals(msg.getStatus())) {
                String err = "ERROR status ! ";
                getLogger().log(Level.SEVERE, err, msg.getError());
                throw new Exception(err, msg.getError()); // timeout
                
            } else {
                try {
                    NormalizedMessage out = null;
                    // NB. maybe necessary to do again full copy with new NormalizedMessageImpl()
                    if (msg.getMessage("OUT") != null) {
                        out = msg.getMessage("OUT");
                    } else if (msg.getFault() != null) {
                        out = msg.getFault();
                    } else {
                        // will be the same as DONE status received
                    }
                    NormalizedMessage copyOut = msg.createMessage();
                    copy(out, copyOut);

                    msg.setStatus(ExchangeStatus.DONE);
                    getChannel().send(msg);
                    return out;
                    
                } catch (Exception e) {
                    String err = "Error while getting response message";
                    getLogger().log(Level.SEVERE, err, e);
                    throw new Exception(err, e); // timeout
                }
            }
        } else {
            String err = "Timeout !";
            getLogger().log(Level.SEVERE, err);
            throw new Exception(err); // timeout
        }
    }
    
    /**
     * Copies a message's content and attachments in another.
     * @param originalNm
     * @param copyNm
     * @throws Exception
     */
    @SuppressWarnings("unchecked")
    protected static void copy(NormalizedMessage originalNm, NormalizedMessage copyNm) throws Exception {
        // setting content, with hardcopy (because if semantic routing...)
        String contentString = SourceHelper.createString(originalNm.getContent());
        copyNm.setContent(SourceHelper.createSource(contentString));
        
        // setting properties
        Set<String> propertyNames = originalNm.getPropertyNames();
        for (String propertyName : propertyNames) {
            copyNm.setProperty(propertyName, originalNm.getProperty(propertyName));
        }
        
        // setting attachments if any
        Set<String> inAttachmentNameSet = originalNm.getAttachmentNames();
        if (inAttachmentNameSet != null) {
            for (String attachmentName : inAttachmentNameSet) {
                DataHandler attDh = originalNm.getAttachment(attachmentName);
                try {
                    copyNm.addAttachment(attachmentName, attDh);
                } catch (MessagingException e) {
                    throw new Exception("Error when attaching file " + attachmentName, e);
                }
            }
        }
    }
    
    /**
     * Copies a message's content and attachments in a MessageExchangeWrapper.
     * @param originalNm
     * @param exchange
     * @throws Exception
     */
    @SuppressWarnings("unchecked")
    protected static void copy(NormalizedMessage originalNm, MessageExchangeWrapper exchange) throws Exception {
        // setting content, with hardcopy (because if semantic routing...)
        String contentString = SourceHelper.createString(originalNm.getContent());
        exchange.setOutMessageContent(SourceHelper.createSource(contentString));
        
        // setting properties
        Set<String> propertyNames = originalNm.getPropertyNames();
        for (String propertyName : propertyNames) {
            exchange.setProperty(propertyName, originalNm.getProperty(propertyName));
        }
        
        // setting attachments if any
        Set<String> inAttachmentNameSet = originalNm.getAttachmentNames();
        if (inAttachmentNameSet != null && !inAttachmentNameSet.isEmpty()) {
            Map<String, DataHandler> attachements = new HashMap<String, DataHandler>();
            for (String attachmentName : inAttachmentNameSet) {
                DataHandler attDh = originalNm.getAttachment(attachmentName);
                attachements.put(attachmentName, attDh);
            }
            exchange.setOutMessageAttachements(attachements);
        }
    }

    /**
     * Create a DOM Node result from an XML Source
     * 
     * @param source the XML Source
     * @return a DOM representation of the XML Document
     * @throws TransformerException
     * @throws IOException 
     * @throws PEtALSComponentSDKException 
     * @throws IOException 
     * @throws PEtALSComponentSDKException 
     */
    public static Node createDOMNodeFromSource(Source source) throws TransformerException, IOException, PEtALSComponentSDKException {
        ((StreamSource) source).getInputStream().reset();
        String sourceString = SourceHelper.createString(source);
        source = SourceHelper.createSource(sourceString);
        
        DOMResult resultDom = new DOMResult();
        TransformerFactory tFactory = TransformerFactory.newInstance();
        Transformer transformer = tFactory.newTransformer();
        transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
        transformer.transform(source, resultDom);
        Node resultNode = resultDom.getNode();
        return resultNode;
    }
    
    /**
     * Return the next sibling of a node, regardless <i>text</i> node
     * @param node
     * @return
     */
    public static Node getNextSibling(final Node node) {
        Node result = node.getNextSibling();
        while (result != null && result.getNodeType() != Node.ELEMENT_NODE) {
            result = result.getNextSibling();
        }
        return result;
    }
    
    /**
     * Find the XPath of a node.
     * @param node The node
     * @return The XPath of the node
     */
    private static String getDebugPath(Node node) {
        StringBuilder path = new StringBuilder();
        
        Node currentNode = node;
        do {
            String nodePath = "/" +  currentNode.getNodeName() + "[" + getPosition(currentNode) + "]";
            path.insert(0, nodePath);
            
            currentNode = currentNode.getParentNode();
            
        } while(currentNode != null && !currentNode.getNodeName().equals("extension"));
        
        return path.toString();
    }
    
    /**
     * Find the position of a node.
     * @param node The node
     * @return The position of the node (-1 if not found).
     */
    private static int getPosition(Node node) {
        int position = -1;
        
        Node parent = node.getParentNode();
        if (parent != null) {
            
            NodeList children = parent.getChildNodes();
            for (int i = 0; i < children.getLength(); i++) {
                Node child = children.item(i);
                
                if (child.getNodeType() != Node.ELEMENT_NODE) {
                    continue;
                }
                
                if (node.getNodeName().equals(child.getNodeName())) {
                    position++;
                }
                
                if (node == child) {
                    return position;
                }
            }
        }

        return -1; // not found
    }
    
    public static void viewNode(Node node) {
        try {
            DOMSource domSource = new DOMSource(node);
            StringWriter writer = new StringWriter();
            StreamResult result = new StreamResult(writer);
            TransformerFactory tf = TransformerFactory.newInstance();
            Transformer transformer = tf.newTransformer();
            transformer.transform(domSource, result);
            String stringResult = writer.toString();
            System.out.println("stringResult:" + stringResult);
        } catch (Exception e) {
            e.printStackTrace();
        }
    }
}
